@@ -15,6 +15,11 @@ from sklearn.metrics import confusion_matrix
 
 from exputils.io import create_filepath
 
+# TODO generalize confusion matrix to multiple discrete RVs with a
+# ConfusionTensor class, which would be the parent to ConfusionMatrix
+
+# TODO consider ways to interlink this with LabelEncoder better. If necessary
+
 
 class ConfusionMatrix(object):
     """Confusion matrix for nominal data that wraps the
@@ -28,7 +33,7 @@ class ConfusionMatrix(object):
         targets : np.ndarray()
             The confusion matrix to wrap, or the target labels.
         preds : np.ndarray, optional
-        labels : list, optional
+        labels : np.ndarray, optional
             The labels for the row and columns of the confusion matrix
         """
         if preds is None:
@@ -41,8 +46,16 @@ class ConfusionMatrix(object):
                 and targets.shape[0] == targets.shape[1]
             ):
                 # If given an existing matrix as a confusion matrix
-                self.labels = labels
                 self.mat = np.array(targets)
+
+                if isinstance(labels, list):
+                    self.labels = np.array(labels)
+                elif isinstance(labels, np.ndarray):
+                    self.labels = labels
+                elif isinstance(targets, pd.DataFrame):
+                    self.labels = np.array(targets.columns)
+                else:
+                    self.labels = np.arange(len(self.mat))
             else:
                 raise TypeError(' '.join([
                     'targets type is expected to be of type `np.ndarray`, but',
@@ -50,20 +63,107 @@ class ConfusionMatrix(object):
                 ]))
         elif preds is not None:
             # Calculate the confusion matrix from targets and preds with sklearn
+            if isinstance(labels, list):
+                self.labels = np.array(labels)
+            elif isinstance(labels, np.ndarray):
+                self.labels = labels
+            else:
+                self.labels = np.array(list(set(targets) | set(preds)))
+
             self.mat = confusion_matrix(
                 targets,
                 preds,
-                labels=labels,
+                labels=self.labels,
                 *args,
                 **kwargs,
             )
 
-            self.labels = labels
+    def __add__(self, other):
+        """Add two ConfusionMatrices together if of the same shape w/ same
+        label set."""
+        if not isinstance(other, ConfusionMatrix):
+            raise TypeError(
+                'Operator add only supported between two ConfusionMatrices.'
+            )
+
+        if len(self.labels) != len(other.labels):
+            raise ValueError(' '.join([
+                'The two ConfusionMatrices do not have the same number of',
+                'labels!',
+            ]))
+        if not self.labels == other.labels:
+            if not set(self.labels) == set(other.labels):
+                raise ValueError(
+                    'The other ConfusionMatrix does not have the same labels!'
+                )
+
+            # TODO reorganize other to be added correctly to existing conf mat
+            raise NotImplementedError('Same labels of different order.')
+
+        return self.mat + other.mat
+
+    def reduce(self, labels, reduced_label, inverse=False):
+        """Reduce confusion matrix to smaller size by mapping labels to one.
+
+        Parameters
+        ----------
+        labels : list | np.ndarray
+            Labels to be reduced to a single label. If `inverse` is True, then
+            all the labels in the confusion matrix NOT included in this list
+            are to be reduced.
+        reduced_label : object
+            The label to replace the others.
+        inverse : bool, optional
+            If True, all of the labels in the confusion matrix NOT in `labels`
+            are reduced instead.
+
+        Returns
+        -------
+        ConfusionMatrix
+            The resulting reduced confusion matrix.
+        """
+        if inverse:
+            if reduced_label in self.labels and reduced_label in labels:
+                labels = np.delete(
+                    labels,
+                    np.where(labels == reduced_label)[0][0],
+                )
+        else:
+            if reduced_label in self.labels and reduced_label not in labels:
+                labels = np.append(labels, reduced_label)
+
+        # TODO the use of the nominal label encoder would be good here.
+        mask = np.zeros(len(self.labels)).astype(bool)
+        for label in labels:
+            mask |= self.labels == label
+
+        if inverse:
+            # Numpy mask in where= ignores False from the calculation.
+            not_mask = mask
+            mask = np.logical_not(mask)
+        else:
+            not_mask = np.logical_not(mask)
+
+        # Construct reduced st reduced label is last index
+        reduced_cm = np.vstack((
+            np.hstack((
+                self.mat[not_mask][:, not_mask],
+                self.mat[not_mask][:, mask].sum(1, keepdims=True),
+            )),
+            np.hstack((
+                self.mat[mask][:, not_mask].sum(0, keepdims=True),
+                self.mat[mask][:, mask].sum(keepdims=True),
+            )),
+        ))
+
+        return ConfusionMatrix(
+            reduced_cm,
+            labels=np.append(self.labels[not_mask], reduced_label),
+        )
 
     # TODO methods for the metrics able to be derived from the confusion matrix
     # f score (1 and beta)
     # informedness # This may not generalize to multi class
-    # mathew's correlation coefficient (MCC)
     # ROC
     # ROC AUC
     # TOC
@@ -80,7 +180,7 @@ class ConfusionMatrix(object):
             raise NotImplementedError('Use sklearn.metrics on the samples')
         return 1.0 - self.accuracy()
 
-    def true_rate(self, average=False, label_weights=None):
+    def true_rate(self, label=None, average=False, label_weights=None):
         """Recall, sensitivity, hit rate, or true positive rate. This is
         calculated the same as specificity, selectivity or true negative rate,
         but on the different classes.
@@ -91,8 +191,15 @@ class ConfusionMatrix(object):
             if True, averages all true class rates together. Otherwise, returns
             the true rate per class in order of labels.
         """
-        recalls = np.diagonal(self.mat) / self.mat.sum(axis=1)
+        #recalls = np.diagonal(self.mat) / self.mat.sum(axis=1)
+        if label is None:
+            tp = self.true_positive()
+            return tp / (tp + self.false_negative())
 
+        tp = self.true_positive(label)
+        return tp / (tp + self.false_negative(label))
+
+        """
         if average:
             if label_weights is not None:
                 raise NotImplementedError('Use sklearn.metrics on the samples')
@@ -101,6 +208,53 @@ class ConfusionMatrix(object):
 
         # Provide the recall per class
         return recalls
+        """
+
+    def true_positive(self, label=None):
+        if label is None:
+            return np.diag(self.mat)
+        return np.diag(self.mat)[label]
+
+    def false_positive(self, label=None):
+        """Calculates the False Positives either of every class or a specified
+        slice.
+
+        Parameters
+        ----------
+        label : np.ndarry slice
+            A slice of the conufsion matrix indices, NOT the labels themselves
+        """
+        # TODO again, support for NominalDataEncoder would make labels able to
+        # support the actual labels
+        if label is None:
+            return self.mat.sum(0) - np.diag(self.mat)
+        return self.mat[label].sum() - self.mat[label, label]
+
+    def false_negative(self, label=None):
+        """Calculates the False Negatives of every class or a specified slice.
+        """
+        if label is None:
+            return self.mat.sum(1) - np.diag(self.mat)
+        mask = np.ones(self.mat.shape[0]).astype(bool)
+        mask[label] = False
+        return self.mat[:, label].sum() - self.mat[label, label]
+
+    def false_rates(self, label):
+        """Calcuate the False Positive/Negative Rates for a single label."""
+        tp = self.true_positive(label)
+        fp = self.false_positive(label)
+        fn = self.false_negative(label)
+        tn = self.mat.sum() - tp - fp - fn
+
+        logging.debug('tp = %f', tp)
+        logging.debug('tn = %f', tn)
+        logging.debug('fp = %f', fp)
+        logging.debug('fn = %f', fn)
+
+        fpr = fp / (fp + tn)
+        fnr = fn / (tp + fn)
+
+        return fpr, fnr
 
     def f_score(self, beta=1, average='macro', label_weights=None):
         # average match sklearn f1_score: None, "binary", micro, macro,
@@ -119,6 +273,9 @@ class ConfusionMatrix(object):
         correct_pred = np.diagonal(self.mat).sum()
         total_sqrd = self.mat.sum()**2
 
+        # TODO be aware that NaNs occur at times, and the below may need a
+        # nonzero mask like mutual information did!
+
         return (
             (correct_pred * self.mat.sum() - np.dot(actual, predicted)) /
             (
@@ -155,6 +312,17 @@ class ConfusionMatrix(object):
         #   thus, it follow geometric mean is similarly used, as is min() and
         #   max()
 
+        # TODO I did the check and they are close but slight difference occurs
+        # at the 16th decimal place and on, which make sense cuz of float
+        # errors, but I am uncertain if they have a workaround to ensure it is
+        # more exact in the scikit-learn code.
+        # Using test example actual = [3, 2, 0, 0, 4] and pred = [4, 2, 3, 0,
+        # 1], The below MI and NMI for min, max, & arithmetic all matched ==,
+        # but geometric did not match exactly. it was off by ~1e-16.
+        # This may be due to not removing zero prob items from  marginals in
+        # entropy calc?
+        # Oh hey, the above is one unit test I can add. LOL
+
         if base is None:
             log = np.log
         elif base == 2:
@@ -166,21 +334,30 @@ class ConfusionMatrix(object):
         marginal_actual = joint_distrib.sum(1).reshape(-1,1)
         marginal_pred = joint_distrib.sum(0).reshape(1,-1)
 
-        # TODO need a unit test for this
+        # TODO need a unit test for this; have confirmed matches
+        # sklearn.metrics.mutual_info_score, tho, when using natural base 'e'
         denom = np.dot(marginal_actual, marginal_pred)
 
+        # Mask nonzero joint to avoid unnecessary zero division
+        nonzero = joint_distrib != 0
+
         if (
             isinstance(weights, np.ndarray)
             and weights.shape == joint_distrib.shape
         ):
             # Weighted MI from Guiasu 1977
             mutual_info = (
-                weights.flatten() * joint_distrib
-                * log(joint_distrib / denom)
+                weights.flatten() * joint_distrib[nonzero]
+                * log(joint_distrib[nonzero] / denom[nonzero])
             ).sum()
         else:
-            mutual_info = (joint_distrib * log(joint_distrib / denom)).sum()
+            mutual_info = (
+                joint_distrib[nonzero]
+                * log(joint_distrib[nonzero] / denom[nonzero])
+            ).sum()
 
+        # TODO able to avoid re-calc of marginals by simply calling scipy
+        # entropy here! Also avoiding two func calls.
         if normalized is None:
             return mutual_info
         if normalized == 'arithmetic':
